/**
 * @file conf.c  Application Configuration
 *
 * Copyright (C) 2010 Creytiv.com
 */
#include <re.h>
#include <fcntl.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <stdio.h>
#include <string.h>
#ifdef HAVE_IO_H
#include <io.h>
#endif
#include <rem.h>
#include <baresip.h>
#include "core.h"


#define DEBUG_MODULE "conf"
#define DEBUG_LEVEL 5
#include <re_dbg.h>


#ifdef WIN32
#define open _open
#define read _read
#define close _close
#endif


#undef MOD_PRE
#define MOD_PRE ""  /**< Module prefix */


static const char *conf_path = NULL;
static char modpath[256] = ".";
static struct conf *conf_obj;
static char dns_domain[64] = "domain";


/** Core Run-time Configuration - populated from config file */
struct conf config = {
	/* Input */
	{
		"/dev/event0",
		5555
	},

	/** SIP User-Agent */
	{
		16,
		""
	},

	/** Audio */
	{
		"","",
		"","",
		"","",
		{8000, 48000},
		{1, 2},
		0,
		{0, 0},
		{0, 0},
		false,
	},

	/** Video */
	{
		"", "",
		352, 288,
		384000,
		25,
		"",
		""
	},

	/** Audio/Video Transport */
	{
		0xb8,
		{10000, 20000},
		{512000, 1024000},
		true,
		false,
		{5, 10}
	},
	SOUND_PORTAUDIO_DSOUND,
	{
		true,
		false,
		false
	}
};


/**
 * Parse a config file, calling handler for each line
 *
 * @param filename Config file
 * @param ch       Line handler
 *
 * @return 0 if success, otherwise errorcode
 */
int conf_parse(const char *filename, confline_h *ch)
{
	struct pl pl, val;
	struct mbuf *mb;
	int err = 0, fd = open(filename, O_RDONLY);
	if (fd < 0)
		return errno;

	mb = mbuf_alloc(1024);
	if (!mb) {
		err = ENOMEM;
		goto out;
	}

	for (;;) {
		uint8_t buf[1024];

		const ssize_t n = read(fd, (void *)buf, sizeof(buf));
		if (n < 0) {
			err = errno;
			break;
		}
		else if (n == 0)
			break;

		err |= mbuf_write_mem(mb, buf, n);
	}

	pl.p = (const char *)mb->buf;
	pl.l = mb->end;

	while (pl.p < ((const char *)mb->buf + mb->end) && !err) {
		const char *lb = pl_strchr(&pl, '\n');

		val.p = pl.p;
		val.l = lb ? (uint32_t)(lb - pl.p) : pl.l;
		pl_advance(&pl, val.l + 1);

		if (!val.l || val.p[0] == '#')
			continue;

		err = ch(&val);
	}

 out:
	mem_deref(mb);
	(void)close(fd);

	return err;
}
 


void conf_path_set(const char *path)
{
	conf_path = path;
}


int conf_path_get(char *path, uint32_t sz)
{
	/* Use explicit conf path */
	if (conf_path) {
		if (re_snprintf(path, sz, "%s", conf_path) < 0)
			return ENOMEM;
		return 0;
	}

	return -1;
}



static int conf_get_range(struct conf *conf, const char *name,
			  struct range *rng)
{
	struct pl r, min, max;
	uint32_t v;
	int err;

	err = conf_get(conf, name, &r);
	if (err)
		return err;

	err = re_regex(r.p, r.l, "[0-9]+-[0-9]+", &min, &max);
	if (err) {
		/* fallback to non-range numeric value */
		err = conf_get_u32(conf, name, &v);
		if (err) {
			DEBUG_WARNING("%s: could not parse range: (%r)\n",
				      name, &r);
			return err;
		}

		rng->min = rng->max = v;

		return err;
	}

	rng->min = pl_u32(&min);
	rng->max = pl_u32(&max);

	return 0;
}


static int get_video_size(struct conf *conf, const char *name,
			  struct vidsz *sz)
{
	struct pl r, w, h;
	int err;

	err = conf_get(conf, name, &r);
	if (err)
		return err;

	w.l = h.l = 0;
	err = re_regex(r.p, r.l, "[0-9]+x[0-9]+", &w, &h);
	if (err)
		return err;

	if (pl_isset(&w) && pl_isset(&h)) {
		sz->w = pl_u32(&w);
		sz->h = pl_u32(&h);
	}

	/* check resolution */
	if (sz->w & 0x1 || sz->h & 0x1) {
		DEBUG_WARNING("video_size should be multiple of 2 (%ux%u)\n",
			      sz->w, sz->h);
		return EINVAL;
	}

	return 0;
}


static int dns_server_handler(const struct pl *pl, void *arg)
{
	struct sa sa;
	int err;

	(void)arg;

	err = sa_decode(&sa, pl->p, pl->l);
	if (err) {
		DEBUG_WARNING("dns_server: could not decode `%r'\n", pl);
		return err;
	}

	err = net_dnssrv_add(&sa);
	if (err) {
		DEBUG_WARNING("failed to add nameserver %r: %s\n", pl,
			      strerror(err));
	}

	return err;
}



#ifdef STATIC

/* Declared in static.c */
extern const struct mod_export *mod_table[];

static const struct mod_export *find_module(const struct pl *pl)
{
	struct pl name;
	uint32_t i;

	if (re_regex(pl->p, pl->l, "[^.]+.[^]*", &name, NULL))
		name = *pl;

	for (i=0; ; i++) {
		const struct mod_export *me = mod_table[i];
		if (!me)
			return NULL;
		if (0 == pl_strcasecmp(&name, me->name))
			return me;
	}
	return NULL;
}
#endif


static int load_module(struct mod **mp, const struct pl *name)
{
	char file[256];
	struct mod *m;
	int err = 0;

	if (!name)
		return EINVAL;

#ifdef STATIC
	/* Try static first */
	err = mod_add(&m, find_module(name));
	if (!err)
		goto out;
#endif

	/* Then dynamic */
	if (re_snprintf(file, sizeof(file), "%s/%r", modpath, name) < 0) {
		err = ENOMEM;
		goto out;
	}
	err = mod_load(&m, file);
	if (err)
		goto out;

 out:
	if (err) {
		DEBUG_WARNING("module %r: %s\n", name, strerror(err));
	}
	else if (mp)
		*mp = m;

	return err;
}


int conf_load_module(struct mod **mp, const char *name)
{
	char buf[64];
	struct pl pl;

	if (re_snprintf(buf, sizeof(buf), MOD_PRE "%s" MOD_EXT, name) < 0)
		return ENOMEM;

	pl_set_str(&pl, buf);

	return load_module(mp, &pl);
}


static int module_handler(const struct pl *pl, void *arg)
{
	(void)arg;

	(void)load_module(NULL, pl);

	return 0;
}


static int config_mod_parse(struct conf *conf)
{
	int err;

	/* Modules */
	if (conf_get_str(conf, "module_path", modpath, sizeof(modpath)))
		str_ncpy(modpath, ".", sizeof(modpath));

	err = conf_apply(conf, "module", module_handler, NULL);
	if (err)
		goto out;

 out:
	if (err) {
		DEBUG_WARNING("configure module parse error (%s)\n",
			      strerror(err));
	}

	return err;
}


int configure(void)
{
	struct pl pollm;
	enum poll_method method;
	uint32_t v;
	int err = 0;
#if 0
	/* Core */
	if (0 == conf_get(conf, "poll_method", &pollm)) {
		if (0 == poll_method_type(&method, &pollm)) {
			err = poll_method_set(method);
			if (err) {
				DEBUG_WARNING("poll method (%r) set: %s\n",
						  &pollm, strerror(err));
			}
		}
		else {
			DEBUG_WARNING("unknown poll method (%r)\n", &pollm);
		}
	}
#endif
	struct pl dummy;
	struct pl modname;
	pl_set_str(&dummy, "");

	//dns_server_handler(&dummy, NULL);

	if (config.codecs.g711)
	{
		pl_set_str(&modname, "g711");
		load_module(NULL, &modname);
	}

	if (config.codecs.g726_32)
	{
		pl_set_str(&modname, "g726_32");
		load_module(NULL, &modname);
	}

	if (config.codecs.gsm)
	{
		pl_set_str(&modname, "gsm");
		load_module(NULL, &modname);
	}

	if (config.codecs.speex)
	{
		pl_set_str(&modname, "speex");
		load_module(NULL, &modname);
	}

	if (config.sound_backend == SOUND_WINWAVE) {
		pl_set_str(&modname, "winwave");
	} else {
		// default: portaudio/DirectSound
		pl_set_str(&modname, "portaudio");
	}
	load_module(NULL, &modname);

	//TODO: AEC seems to not work well
#if 0
	pl_set_str(&modname, "speex_aec");
	load_module(NULL, &modname);
#endif

	pl_set_str(&modname, "stun");
	load_module(NULL, &modname);


	return err;
}



/**
 * Get the path to plugin-modules
 *
 * @return Path to modules
 */
const char *conf_modpath(void)
{
	return modpath;
}


/**
 * Get the current configuration object
 *
 * @return Config object
 *
 * @note It is only available during init
 */
struct conf *conf_cur(void)
{
	return conf_obj;
}


/**
 * Set the DNS domain for config templates
 *
 * @param domain DNS domain
 */
void conf_set_domain(const char *domain)
{
	str_ncpy(dns_domain, domain, sizeof(dns_domain));
}
